import logging
import numpy as np
import pandas as pd
import re

# divide between term (r) and relation (r) in relation path
divider_tr = '.'
divider_rt = '~'


def _tuple_list_to_dict(tuple_list):
    tuple_dict = {}
    for k, v in tuple_list:
        try:
            tuple_dict[k].append(v)
        except KeyError:
            tuple_dict[k] = [v]
    return tuple_dict


def relation_path_to_text(relation_path, ont):
    """
    Converts from a relation string e.g. "UBERON:123913.is_a~UBERON:1381239" to a text version,
     e.g. "heart is a circulatory organ".

    :param ont: opy.Obo() ontology object.
    :param relation_path: path describing the relationship between two terms, e.g. "UBERON:123913.is_a-UBERON:1381239"
    :return:
    """
    if pd.isna(relation_path):
        return relation_path
    # TODO: add fix for alt-ids, obsolete terms, etc. For now just keep their ID.
    for i, sub_relation in enumerate(relation_path.split(divider_tr)):
        if i == 0:
            # sub_relation is actually the source *term* identifier for i == 0.
            try:
                relation_text = ont[sub_relation]['name']
            except KeyError:
                relation_text = sub_relation
            continue
        relation = sub_relation.split(divider_rt)[0].replace('_', ' ')
        term_id = sub_relation.split(divider_rt)[-1]
        try:
            relation_text += f" {relation} {ont[term_id]['name']}"
        except KeyError:
            relation_text += f" {relation} {term_id}"
    return relation_text


def _found_term(relation_path):
    """
    Finds the last term in the relation path.

    :param relation_path:
    :return:
    """

    if pd.isna(relation_path):
        return relation_path
    else:
        return relation_path.split(divider_rt)[-1]


def _check_if_found(new_term: str, targets: list) -> bool:
    """
    Checks if `new_term` matches `targets`.

    :param new_term: string to check
    :param targets: list of strings to match to `new_term`. Targets can be a list of specific targets, e.g.
        ['UBERON:123219', 'UBERON:1288990'] or of general ontology prefixes, e.g. ['UBERON']
    :return:
    """

    if (':' in targets[0]) and (new_term in targets):  # specific
        relation_found = True
    elif (':' not in targets[0]) and (new_term.split(':')[0] in targets):  # general
        relation_found = True
    else:
        relation_found = False

    return relation_found


class Relations(pd.DataFrame):

    def __init__(self, allowed_relations: list, ont, sources=None, targets=None, source_targets=None, excluded=None, col_names=None, mode='any'):
        """
        Pandas Dataframe containing relationships between `sources` and `targets` terms according to `ont`.
        Finds relationships that do not pass through `excluded` terms and uses only `allowed_relations`. We keep looking
        until we find a relation to a target (if mode == 'any') or we run out of leads.

        :param allowed_relations: a list of allowed relations, e.g. ['is_a', 'part_of']
        :param sources: list of sources. For mode `all` must be a list of source-target tuple airs.
        :param mode: 'any' or 'all' - 'all is looking for specific term1-term2 pairs, while 'any' is looking for any
          relationship between something in specific source and anything in targets.
        :param targets: list of targets.
        :param source_targets: list of tuples of source-target pairs. Do not provide source or targets if using this
          parameter. Only runs in "all" mode.
        :param ont: Obo ontology object.
        :param excluded: a list/set of terms which are explicitly not being searched for (which may otherwise match the
          targets). Useful e.g. if we want to look for any tissue targets with prefix 'UBERON', except for very general
          ones. Does not allow relationships that pass through this term.
        :param col_names: Alternative column names for the output of Relations Data Frame, by default is
          ['from', 'relation_path', 'relation_text', 'to']
        """
        # TODO: Add default for allowed_relations?
        # TODO: put parameters in order

        assert (mode in ['any', 'all'])

        if source_targets:
            assert mode == 'all'
            assert (not sources)
            assert (not targets)
            assert (not excluded)
        else:
            assert targets
            assert isinstance(targets, list)

        if col_names is None:
            col_names = ['from', 'relation_path', 'relation_text', 'to']
        else:
            assert len(col_names) == 4

        if excluded is None:
            excluded = []

        assert(isinstance(allowed_relations, list))

        if not isinstance(sources, type(None)):
            if mode == 'any':
                assert sources.__iter__

        # TODO: remove/test source_targets.
        #
        #     else:
        #         source_targets = zip(sources, targets)
        #
        # if source_targets:
        #     assert source_targets.__iter__
        #     row_index = source_targets

        super(Relations, self).__init__(data=None,
                                        index=sources,
                                        columns=col_names[1:],
                                        dtype=None,
                                        copy=True)
        self.index.rename(col_names[0], inplace=True)

        if mode == 'any':
            self._calculate_any(allowed_relations, targets, ont, excluded)
        elif mode == 'all':
            # TODO: fix/test for both source-target and source-and-target modes
            self._calculate_all(allowed_relations, targets, ont, excluded)

    def _calculate_all(self, allowed_relations, targets, ont, excluded):
        """
        Looks for relations between all specified pairs of source term to target term.

        Basically, only stops looking when we stop getting new relations.

        :param allowed_relations:
        :param ont:
        :param excluded:
        :return:
        """
        # TODO: Add functionaltiy for source_targets, or remove because this function is the same as _calculate_any
        found_relation_paths = []
        for source in self.index:
            found_relation_path_list = _find_relation(source, allowed_relations, targets, ont, excluded, 'all')
            found_relation_paths.append(found_relation_path_list)

        # Format output:
        self.iloc[:, 0] = found_relation_paths
        self.iloc[:, 1] = [[relation_path_to_text(pth, ont) for pth in lst] for lst in found_relation_paths]
        self.iloc[:, 2] = [[_found_term(pth) for pth in lst] for lst in found_relation_paths]

    def _calculate_any(self, allowed_relations, targets, ont, excluded):
        """
        Looks for relation of any souce term to any target term. Stops looking when relation found.

        :param allowed_relations:
        :param targets:
        :param ont:
        :param excluded:
        :return:
        """
        found_relation_paths = []
        for source in self.index:
            # TODO: Make sure that _check_if_found can handle both types of target
            found_relation_path = _find_relation(source, allowed_relations, targets, ont, excluded)
            found_relation_paths.append(found_relation_path)

            # Format output:
        self.iloc[:, 0] = found_relation_paths
        self.iloc[:, 1] = [relation_path_to_text(relation_path, ont) for relation_path in found_relation_paths]
        self.iloc[:, 2] = [_found_term(relation_path) for relation_path in found_relation_paths]

    def format_all(self, ont, targets):
        """
        Creates a nicely formatted multi-indexed DataFrame with (source, target) pairs. Useful when using "mode=all".

        :param ont: ontology to look up paths.
        :param targets: list of term IDs, e.g. ['GO', 'UBERON']
        :return:
        """
        # Get all source-target pairs (keys) and relation paths/strings (list as values)
        source_target_dict = {}
        for tissue_term, row in self.iterrows():
            for i, relation_path in enumerate(row['relation_path']):
                for target in targets:
                    for go_term in re.findall(f'({target}:\d+)', relation_path):
                        relation_path_target = relation_path.split(go_term)[0] + go_term
                        relation_text_target = relation_path_to_text(relation_path_target, ont)
                        source_target_dict[(tissue_term, go_term)] = [relation_path_target, relation_text_target]

        # Format into a multi-indexed data frame:
        formatted_df = pd.DataFrame.from_dict(source_target_dict,
                                              orient='index',
                                              columns=['relation_path', 'relation_text'])
        formatted_df.index = pd.MultiIndex.from_tuples(formatted_df.index, names=["from", "to"])

        return formatted_df


def _find_relation(source, allowed_relations, targets, ont, excluded, mode='any'):
    """
    Searches ontology `ont` for a relationship path between `source` and `target` (self.index), which does not pass
    through `excluded` and uses only `allowed_relations`.

    We keep looking until we find a relation to a target (relation_found == True) or we run out of leads
    (unchanged == True, i.e. we have no new relation strings after another loop).

    :param allowed_relations:
    :param targets: list of types of targets, e.g. ["GO"]
    :param ont:
    :param excluded:
    :return:
    """
    # TODO: check if it's quicker to have a list of tuples ('is_a', 'termID') instead of long string.
    #
    if mode == 'all':
        found_relation_paths = set()

    checked_terms = set()

    relation_paths = [source]
    relation_found = False
    unchanged = False

    while (not (relation_found and mode == 'any')) and (not unchanged):
        new_relation_paths = []
        for relation_path in relation_paths:
            most_recent_term = relation_path.split(divider_rt)[-1]

            if most_recent_term in checked_terms:
                continue
            else:
                checked_terms.add(most_recent_term)

            # Ontologies can contain external terms, e.g. `NCBITaxon:9606`
            try:
                relations = ont[most_recent_term].keys()
            except KeyError:
                relations = []

            for relation in relations:
                if relation not in allowed_relations:
                    new_terms = []
                else:
                    new_terms = ont[most_recent_term][relation]

                # For each new term, check for wanted relation:
                for new_term in new_terms:

                    if new_term in excluded:
                        continue

                    if new_term in relation_path:
                        logging.info(f'cyclic relationship: '
                                     f'{relation_path}{divider_tr}{relation}{divider_rt}{new_term}')
                        continue

                    new_relation_path = f'{relation_path}{divider_tr}{relation}{divider_rt}{new_term}'
                    new_relation_paths.append(new_relation_path)

                    relation_found = _check_if_found(new_term, targets)
                    if relation_found and mode == 'all':
                        found_relation_paths.add(new_relation_path)
                    elif relation_found and mode == 'any':
                        return new_relation_path

        if len(new_relation_paths) == 0:
            unchanged = True
        relation_paths = new_relation_paths

    if mode == 'any':
        return np.nan
    if mode == 'all':
        return found_relation_paths




"""
This module contains code for creating and working with the Obo class: objects that represent ontologies.
"""

import pandas as pd
import urllib.request as request
import os
import logging
import validators


def download_obo(data_name, out_dir='../data/'):
    """
    Download obo from a list of known locations.

    :param data_name: Name of OBO you wish to download.
    :param out_dir: Directory in which to save OBO file.
    :return out_file: path to saved file.
    """
    # TODO: Allow overwrite existing file
    # TODO: test

    uberon_urls = {
    'sensory-minimal':'http://ontologies.berkeleybop.org/uberon/subsets/sensory-minimal.obo',
    'uberon-extended': 'http://purl.obolibrary.org/obo/uberon/ext.obo',
    'uberon-basic': 'http://purl.obolibrary.org/obo/uberon.obo',
    }

    if data_name not in uberon_urls.keys():
        supported_list = '\n'.join(uberon_urls.keys())
        logging.error(f'`data_name` {data_name} not in supported list:\n{supported_list}')

    url = uberon_urls[data_name]
    file_name = (os.path.basename(url))
    out_file = os.path.join(out_dir, file_name)

    if not os.path.isdir(out_dir):
        os.mkdir(out_dir)
        logging.info(f'Created directory {out_dir}.')

    if os.path.isfile(out_file):
        logging.warning(f'File already exists at location: {os.path.abspath(out_file)}. Cancelling download.')
        return out_file

    # TODO: make sure behaviour is sensible if url doesn't exist. (Write test).
    file_data = request.urlopen(url) 
    data_to_write = file_data.read() 
    logging.info(f'Downloaded {data_name} from: {url}')

    with open(out_file, 'wb') as f:
        f.write(data_to_write)

    logging.info(f'Wrote {data_name} file to: {os.path.abspath(out_file)}')
    return out_file


def _validate_term(term_text: str, allowed_ids: list):
    """
    Check term is a valid format.

    :param term_text: text to validate, e.g. "NCBITaxon:9606" or "UBERON:12391203".
    :param allowed_ids: a list of allowed identifier prefixes, e.g. ['GO', 'UBERON'] or None if dont' want to restrict.
    :return: bool True=valid, False=not
    """
    assert(isinstance(term_text, str))
    if not allowed_ids:
        assert(isinstance(allowed_ids, list))
    if ':' not in term_text:
        return False
    elif (allowed_ids and (term_text.split(':')[0] in allowed_ids)):
        return True


def _find_all(haystack, needle):
    idx = -1
    while True:
        idx = haystack.find(needle, idx + 1)
        if idx == -1:
            break
        yield idx


def _find(s: str, ch: str) -> list:
    """
    Finds indices of character `ch` in string `s`
    """
    return [i for i, ltr in enumerate(s) if ltr == ch]


def _between_chars(s: str, ch1: str, ch2=None) -> str:
    if ch2 is None:
        # i1, i2 = _find(s, ch1)[:2]
        i1, i2 = [x for x in _find_all(s, ch1)][:2]
    else:
        i1 = s.find(ch1)
        i2 = s.find(ch2)
    return s[i1 + 1: i2]


def _extract_synonym(rest_of_line: str) -> str:
    """
    Extracts synonym name from obo line.
    :param rest_of_line: string containing line except for "synonym: "
    :return:
    """
    synonym = rest_of_line.split('"')[1].lower()
    return synonym


def _extract_synonym_type(rest_of_line:str) -> str:
    _, i1 = [x for x in _find_all(rest_of_line, '"')][:2]
    i2 = rest_of_line.find('[')
    return rest_of_line[i1+1: i2]


def _read_line_obo(line_list: list, ont_ids: list):
    """
    Reads a line of an obo file, and returns a list of (relation, value) tuples. Most lines will only contain one.
    :param line_list: list of line elements (original line split by spaces)
    :param ont_ids: allowed ontology ids (list): None if want to keep all IDs.
    :return new_relations: a list of (relation, value) tuples
    """

    if not ont_ids:
        assert(isinstance(ont_ids, list))
    assert(isinstance(line_list, list))

    line_list[0] = line_list[0].replace(':', '')
    new_relations = []  # list of (relation, value) tuples.

    if line_list[0] in Obo._text_attributes:
        new_relations.append((line_list[0], ' '.join(line_list[1:])))

    elif line_list[0] in Obo._attributes:
        new_relations.append((line_list[0], line_list[1]))

    elif line_list[0] in Obo._relationships:
        new_relations.append((line_list[0], line_list[1]))

    elif line_list[0] in Obo._nestable_attributes:
        if line_list[1] in Obo._relationships:
            new_relations.append((line_list[1], line_list[2]))
        elif ':' in line_list[1]:
            new_relations.append((line_list[0], line_list[1]))
        elif ':' in line_list[2]:
            # TODO: Add test
            logging.info(f'Relationship {line_list[1]} is not currently stored.'
                         f' saving as {line_list[0]}. {line_list[2]}.')
            new_relations.append((line_list[0], line_list[2]))
        else:
            logging.warning(f'Line looks unusual: {" ".join(line_list)}')

    elif line_list[0] == 'synonym':
        rest = ' '.join(line_list[1:])  # rest of line

        source_relations = _extract_source(rest, ont_ids)
        new_relations += source_relations

        new_relations.append((line_list[0], rest))

    elif line_list[0] == 'def':
        # TODO: update to contain
        rest = ' '.join(line_list[1:])
        new_relations.append((line_list[0], rest))
        new_relations += _extract_source(rest, ont_ids)

    elif line_list[0] == 'xref':
        if _validate_term(line_list[1], ont_ids):
            new_relations.append((line_list[0], line_list[1]))

    # TODO: add logging info for unrecognised lines

    return new_relations


def _extract_source(text, ont_ids):
    """
    In obo files sources are written between square brackets. Currently does not keep URL sources (e.g. wikipedia).

    :param text:
    :param ont_ids: list of allowed ontology ids , e.g. ['GO', 'HP'], or None if don't want to restrict.
    :return new_relations: list of (relation, value) tuples, e.g. [('xref': 'HP:091231')]
    # TODO: What kind of relations should sources be? 'xref'? My own, e.g. 'source'/'source.synonym'/'source.GO.synonym'
    """
    new_relations = []
    sources = [x.strip() for x in _between_chars(text, '[', ']').split(',')]
    for source in sources:
        if _validate_term(source, ont_ids):
            # TODO: explain in Ontolopy part
            new_relations.append((source.split(':')[0], source))
        elif validators.url(source):
            new_relations.append(('url', source))
    return new_relations


def _merge_dict(a, b, prefer='self', path=None):
    """
    Recursively merges dictionary a into dictionary b. Prefers a.

    :param a: dictionary a (self)
    :param b: dictionary b (new)
    :param path: used in recursion
    :return:
    """
    c = a.copy()
    if path is None:
        path = []
    for key in b:
        if key in c:
            if isinstance(c[key], dict) and isinstance(b[key], dict):
                _merge_dict(c[key], b[key], path + [str(key)])
            elif c[key] == b[key]:
                pass  # same leaf value
            elif isinstance(c[key], list) and isinstance(b[key], list):
                c[key] = list(set(b[key]) | set(c[key]))
            else:
                if key == 'namespace':
                    c[key] = f"Combined {c[key]} and {b[key]}"
                elif key == 'name':
                    # we expect some differences in name, due to e.g. terms becoming obsolete
                    logging.info(f"Unmatching names: {c[key]} =/= {b[key]}, keeping the former.")
                else:
                    logging.error(f"For key {key}, {c[key]} =/= {b[key]}. Choosing {prefer}.")
                    if prefer == 'new':
                        c[key] = b[key]
        else:
            c[key] = b[key]
    assert (len(c) == (len(a) + len(b) - len(set(a.keys()) & set(b.keys()))))
    return c


def load_obo(file_loc, ont_ids=None, discard_obsolete=True):
    """
    Loads ontology from `.obo` file at `file_loc`.

    :param file_loc: file location - path to stored obo file.
    :param ont_ids: list of ontology ids, e.g. `['UBERON', 'CL']`
    :param discard_obsolete: if True discard obsolete terms.
    :return: `Obo` ontology object.
    """
    obo = Obo()
    # TODO: check for version/date of ontology file and save if possible
    # terms = {}
    if not ont_ids:
        assert(isinstance(ont_ids, list))
    with open(file_loc) as f:
        term = {}
        for i, line in enumerate(f):
            line = line.strip()
            line = line.strip().split(' ')

            if '[Term]' in line[0]:
                if len(term) > 0 and 'id' in term.keys():
                    if 'comment' in term.keys() and 'obsolete' in term['comment'].lower() and discard_obsolete:
                        logging.info(f"term {term['id']}: {term['name']} is obsolete. Discarding.")
                        term = {}
                        continue
                    elif (not ont_ids) or (ont_ids and (term['id'].split(':')[0] in ont_ids)):
                        obo[term['id']] = term
                term = {}

            new_relations = _read_line_obo(line, ont_ids)
            for (relation, value) in new_relations:
                if relation in Obo._strings:
                    term[relation] = value
                    continue
                try:
                    term[relation].append(value)
                except KeyError:
                    term[relation] = [value]

    return obo


class Obo(dict):
    """
    Creates `Obo` ontology object from `dict` with ontology terms for keys, mapping to term attributes and relations.

    Each key/term is a dictionary with key: value pairs mapping either:

    1. Attribute (`str`) to value (`str`), e.g. `'name': 'scapula'`

    2. Type of relationship (`str`) to term identifiers (`list`), e.g. `'is_a': ['UBERON:0002513']`

    Info: Obo stands for Open Biological Ontology: a popular file format for building biological ontologies.
    """

    # TODO: tidy so that all attributes are organised in a dict with type (string, list), acceptable formats, nestable
    _strings = [  # must be strings not lists
        'name',
        'id'
    ]

    _nestable_attributes = [
        'relationship',
        'intersection_of',
    ]

    _text_attributes = [
        'name',
        'comments',
    ]

    _attributes = [
        'id',
        'alt_id',
        'subset',
        'is_obsolete',  # TODO: add test
        'replaced_by',  # TODO: add test
        'namespace',
        'consider',
    ]

    _relationships = [
        'is_a',
        'union_of',
        'derives_from',
        'is_model_for',
        'develops_from',
        'part_of',
        'never_in_taxon',
        'present_in_taxon',
        'only_in_taxon',
        'dubious_for_taxon',
        # TODO: test from here down:
        'capable_of',
        'has_part',
        'channel_for',
        'capable_of_part_of',
        'positively_regulates',
        'negatively_regulates',
        'regulates',
        'connects',
        'attaches_to',
        'adjacent_to',
        'drains',
        'supplies',
        'composed_primarily_of',
        'skeleton_of',
        'developmentally_replaces',
        'located_in',
        'produced_by',
        'extends_fibers_into',
        'has_potential_to_develop_into',
        'continuous_with',
    ]

    def __init__(self, source_dict=dict()):
        """
        Initialise self from a source dictionary.

        :param source_dict: `dict` mapping terms to their attributes and relationships.
        """
        # Note: when adding attributes consider how this will effect self.merge()
        assert(isinstance(source_dict, dict))
        self._from_dict(source_dict)

    def __copy__(self):
        copy = Obo(dict(self).copy())

        for att_key, att_val in self.__dict__.items():
            copy.__dict__[att_key] = att_val

        return copy

    # TODO: Write __deepcopy__

    @property
    def terms(self):
        """
        The ontology terms (a `dict_keys` object).
        :return:
        """
        return self.keys()

    @property
    def leaves(self):
        """
        Leaf terms are the most specific terms in the ontology; they have no children, only parents (a `set` object).
        :return:
        """
        return self._get_leaves()

    # TODO: Write roots(), get_roots()

    def _from_dict(self, source_dict):
        """
        Create Obo() from a Python dict.
        :param source_dict: dictionary object from which to create `Obo`
        :return:
        """
        assert(isinstance(source_dict, dict))
        for term, value in source_dict.items():
            self[term] = value
        return self

    def terms_from(self, ont_id: list):
        """
        Returns a set of terms in Obo ontology with prefix in list ont_id
        :param ont_id: list of ontology prefixes e.g. ['HP', 'GO']
        :return:
        """
        return {x for x in self.terms if x.split(':')[0] in ont_id}
    
    # def get_relations(self, relations_of_interest, source_terms, target_term, ont):
    #     """
    #     get_relations finds all relationships (based on relations_of_interest e.g. ['is_a']) between terms like source
    #     _term and terms like target_term, e.g. source_term is_a target_term. Returns a mapping between term and most
    #     specific (least number of steps) relationstring (if one exists, else NaN) for each relevant term in the
    #     ontology.
    #
    #     Args:
    #         relations_of_interest: a list of relations that are relevant for finding relationship between the source
    #         and target term, e.g. ['is_a','is_model_for']
    #         source_terms: terms that we wish to look for relations from, list of the form [source_term_1,
    #         source_term_2, etc] such that we wish to find relationships of the form "source_term is_a target_term" or
    #         "source_term is_a other_term is_a target_term".
    #         target_term: term that we wish to look for relations to, e.g. source_term is_a target_term. Term string,
    #         either specific (e.g. 'FF:0000001') or general (e.g. 'FF')
    #
    #     Returns:
    #         Relations class object, containing relations (see relations class for details), relations_of_interest,
    #         source_term and target_term
    #
    #     """
    #     return relations.Relations(relations_of_interest, source_terms, target_term, ont)

    def merge(self, new, prefer='self'):
        """
        Recursively merges `new` into `self` and returns a merged `Obo` ontology.

        :param new: Obo object (or list of objects) to add.
        :param prefer: prefer 'self' (base Obo) or 'new' (new Obo)
        :return merged: A merged Obo
        """

        if isinstance(new, list):
            # create merged ontology
            merged = new[0].copy()
            for obo in new[1:]:
                merged = self.merge(obo)
            return merged

        # TODO: Make merge work for an arbitrary number of ontologies (must be ordered by preference)
        prefer_options = ['self', 'new']
        try:
            assert(prefer in prefer_options)
        except AssertionError:
            logging.error(f"`prefer` must be in {str(prefer_options)}, not {prefer}.")

        merged = _merge_dict(self, new, prefer)
        return Obo(merged)

    def _get_leaves(self, term_types=None, relations_of_interest=None):
        """
        Get the leaf terms from the ontology only.

        :param term_types: if given, a list of term types to restrict to (e.g. ['UBERON', 'GO']
        :param relations_of_interest: if given, a list of relation types to restrict to (e.g. 'is_a', 'part_of')
        :return:
        """
        if not relations_of_interest:
            relations_of_interest = self._relationships + self._nestable_attributes
        else:
            assert(isinstance(relations_of_interest, list))

        leaves = set(self.terms)
        if term_types is not None:
            assert (isinstance(term_types, list))
            assert (
                all([':' not in x for x in term_types]))  # don't want 'UBERON:1231239' must be of form 'UBERON', 'GO'
            leaves = {x for x in leaves if x.split(':')[0] in term_types}

        for term in self.terms:
            for relation in self[term].keys():
                if relation not in relations_of_interest:
                    continue
                related_terms = self[term][relation]
                assert(isinstance(related_terms, list))
                leaves -= set(related_terms)

        return leaves

    # TODO: Write to_json()



"""
This module contains code for creating and working with the Uberon class: ontology containing Uberon terms.
"""
import logging
import pandas as pd
import numpy as np

from .obo import Obo, _extract_synonym, _extract_synonym_type
from .relations import Relations

# TODO: Properly consider this architecture. Seems a bit weird :/.


def uberon_from_obo(obo):
    """
    Creates an `Uberon` object from an `Obo` object.
    Uberon objects have additional functions they can use.

    :param obo: `Obo` object
    :return uberon: `Uberon` object
    """
    uberon = Uberon()
    for term, value in obo.items():
        uberon[term] = value

    for att_key, att_val in obo.__dict__.items():
        uberon.__dict__[att_key] = att_val

    uberon._check_terms()

    return uberon


class Uberon(Obo):
    """
    An UBERON-specific ontology object.
    """

    def __init__(self):
        self._check_terms()

    def _check_terms(self):
        if len(self.terms) == 0:
            return None

        try:
            assert('UBERON' in {x.split(':')[0] for x in self.terms})
        except AssertionError:
            logging.error(f"There are no UBERON terms in your Uberon object.")

    def sample_map_by_ont(self, sample_ids: list, exclude=None, relation_types=None, to=None, child_mapping=False):
        """
        Map tissues from sample names to uberon identifiers. Will only work if ontology contains Uberon + Sample terms.

        :param sample_ids: list of sample identifiers
        :param exclude: list of tissues to exclude, i.e. because they are too general.
        :param relation_types: list of relation types in ontology that relate to position in body.
        :param to: list of ontology prefixes that you want to map to.
        :param child_mapping: If True, searches children instead of parents.
        :return:
        """

        # TODO: add child_mapping functionality
        if exclude is None:
            exclude = [
                'UBERON:0000061',  # anatomical structure
                'UBERON:0000479',  # tissue
                'UBERON:0000467',  # anatomical system
                'UBERON:0011216',  # organ system subdivision
                'UBERON:0000922',  # embryo
                'UBERON:0007023',  # adult organism
                'CL:0000048',  # multi fate stem cell
            ]
        else:
            assert(isinstance(exclude, list))

        if relation_types is None:
            relation_types = [
                'is_a',
                'related_to',
                'part_of',
                'derives_from',
                'intersection_of',
                'union_of',
                # 'is_model_for',
                'replaced_by',
                # 'develops_from',
            ]
        else:
            assert(isinstance(relation_types, list))

        if to is None:
            to = ['UBERON']
        else:
            assert(isinstance(to, list))

        tissue_relations = Relations(
            allowed_relations=relation_types,
            sources=sample_ids,
            targets=to,
            ont=self,
            excluded=exclude,
        )

        return tissue_relations

    def sample_map_by_name(self, sample_names, to=None, col_names=None, xref=None):
        """
        Map tissues from sample identifiers to uberon identifers.

        :param sample_names: map from sample identifiers to tissue/sample descriptors/names for values.
          May be `dict` or `pd.Series`
        :param to: list of ontology prefixes that you want to map to.
        :param xref: An ontology identifier (e.g. FMA) the presence of which denotes a preferred term.
        :param col_names: Column names of returned relationships
        :return:
        """

        # TODO: Currently mapping each name once for each sample... Should just get a a unique name list and go through that.

        # TODO: Make more general (like Relations) and move to Obo() as might want to look at other types of matched
        #  names such as phenotype.

        # TODO: Add terms to exclude
        if col_names is None:
            col_names = ['from', 'name_matched_on', 'to']
        else:
            assert(len(col_names) == 3)

        if to is None:
            to = ['UBERON']
        else:
            assert(isinstance(to, list))

        if isinstance(sample_names, dict):
            sample_names = pd.Series(sample_names)
        assert(isinstance(sample_names, pd.Series))

        name_to_uberon = {}
        for tissue_name in sample_names.unique():
            found = False
            options = []
            if pd.isna(tissue_name):
                name_to_uberon[tissue_name] = np.nan
                continue

            tissue_name_l = tissue_name.lower()
            for uberon_term in self.terms:

                if uberon_term.split(':')[0] not in to:
                    continue

                if self[uberon_term]['name'].lower() == tissue_name_l:
                    name_to_uberon[tissue_name] = uberon_term
                    found = True
                    break

                try:
                    synonyms_info = self[uberon_term]['synonym']
                except KeyError:
                    continue

                # TODO: Allow change allowed syn types
                for synonym_info in synonyms_info:
                    syn_type = _extract_synonym_type(synonym_info)
                    if syn_type not in ['EXACT', 'BROAD', 'NARROW']:
                        continue
                    synonym = _extract_synonym(synonym_info)

                    if tissue_name_l == synonym:
                        # TODO: Check Taxon requirements here
                        options.append((tissue_name, uberon_term, syn_type))
                        found = True

            if not found:
                name_to_uberon[tissue_name] = np.nan
            elif len(options) == 1:
                tissue_name, uberon_term, _ = options[0]
                name_to_uberon[tissue_name] = uberon_term
            elif len(options) >= 1:  # Choose best option of synonyms:
                chosen = False

                # Check for XREF to preferred ontology (e.g. FMA for human)
                if xref is not None:
                    for option in options:
                        tissue_name, uberon_term, _ = option
                        if xref in self[uberon_term].keys():
                            chosen = option

                if not chosen:
                    # TODO: If no XREF choose most specific/most exact synonym? (Currently just pick first)
                    chosen = options[0]
                tissue_name, uberon_term, _ = chosen
                name_to_uberon[tissue_name] = uberon_term

        # Create sample_to_uberon
        sample_to_uberon = pd.DataFrame(columns=col_names[1:], index=sample_names.keys())
        sample_to_uberon.index.rename(col_names[0], inplace=True)
        sample_to_uberon[col_names[1]] = sample_names
        sample_to_uberon[col_names[2]] = sample_names.map(name_to_uberon)

        return sample_to_uberon

    def get_overall_tissue_mappings(self, map_by_name, map_by_ont, rel=None):
        """
        Combines the two mappings `map_by_name` and `map_by_ont` to create an overall mapping and disagreements.

        :param map_by_name: mapping from sample to tissue via sample name, from `Uberon.sample_map_by_name`.
        :type map_by_name: class: `pd.DataFrame`
        :param map_by_ont: mapping from sample to tissue via sample ontology ID, from `Uberon.sample_map_by_ont`.
        :type map_by_ont: class: `pd.DataFrame`
        :parm rel: list of relation strings allowed between name and ontology mappings to count as not a disagreement.
        :type rel: `list`.
        :return: (`overall_mapping`: mapping from sample to tissue combining both sources,
          `disagreements`: disagreements between "by name" and "by ontology" mappings.)
        :rtype: (class: `pd.DataFrame`, class: `pd.DataFrame`)
        """

        samples = list(map_by_ont.index)
        assert(isinstance(map_by_name, pd.DataFrame))
        assert(isinstance(map_by_ont, pd.DataFrame))
        assert(set(map_by_ont.index) == set(map_by_name.index))

        if rel is None:
            rel = Obo._relationships
        assert(isinstance(rel, list))

        map_by_name = map_by_name[~map_by_name.index.duplicated(keep='first')]
        map_by_ont = map_by_ont[~map_by_ont.index.duplicated(keep='first')]

        assert(len(map_by_ont.index) == len(map_by_name.index))

        # Preparing outputs
        mapping_columns = ['sample_id', 'by_name', 'by_ont', 'overall', 'mapped_by']
        overall_mapping = pd.DataFrame(index=map_by_name.index, columns=mapping_columns[1:])
        overall_mapping.index.rename(mapping_columns[0], inplace=True)

        disagreement_columns = ['sample_id', 'by_name', 'by_ont', 'relation_between', 'name_label']
        disagreements = pd.DataFrame(columns=disagreement_columns[1:])
        disagreements.index.rename(disagreement_columns[0], inplace=True)

        for sample in samples:
            # check if mappable by name:
            name_matched_on, by_name = map_by_name.loc[sample]

            # check if mappable by ontology:
            relation_text, by_ont = map_by_ont.loc[sample][1:]

            if not pd.isna(by_name) and pd.isna(by_ont):  # by name only
                overall = by_name
                mapped_using = "name"
            elif not pd.isna(by_ont) and pd.isna(by_name):  # by ont only
                overall = by_ont
                mapped_using = "ontology"
            elif pd.isna(by_name) and pd.isna(by_ont):  # both unmappable
                overall = np.nan
                mapped_using = np.nan
            elif by_name == by_ont:  # both the same
                overall = by_ont
                mapped_using = "both (same)"
            elif by_name != by_ont:  # mappable but different
                name_to_ont = Relations(allowed_relations=rel,
                                        sources=[by_name],
                                        targets=[by_ont],
                                        ont=self).loc[by_name, 'to']
                ont_to_name = Relations(allowed_relations=rel,
                                        sources=[by_ont],
                                        targets=[by_name],
                                        ont=self).loc[by_ont, 'to']

                if pd.isna(name_to_ont) and pd.isna(ont_to_name):  # no relation between by_ont and by_name
                    disagreements.loc[sample] = [by_name, by_ont, relation_text, name_matched_on]
                    overall = by_name
                    mapped_using = "name"
                elif pd.isna(name_to_ont) and not pd.isna(ont_to_name):  # If one is (part of) another:
                    overall = by_name
                    mapped_using = "name"
                    logging.info("name", ont_to_name)
                elif not pd.isna(name_to_ont) and pd.isna(ont_to_name):
                    overall = by_ont
                    mapped_using = "ontology"
                    logging.info("ont", name_to_ont)
            overall_mapping.loc[sample] = [by_name, by_ont, overall, mapped_using]

        return overall_mapping, disagreements

    # TODO: Write restrict_to_taxon, would need to read in NCBI ontology, since it is hierarchical (e.g. vertebrates)
    # def restrict_to_taxon(ont, ncbi_id, valid_terms = ['UBERON'], keep_dubious=False, only_yes=False):
    #     """
    #     Loops through valid_terms terms in ontology `ont`, and removes any that are never in taxon with ID `ncbi_id`
    #     """
    #     yes_relations = ['only_in_taxon', 'present_in_taxon']
    #     never = 'never_in_taxon'
    #     dubious = 'dubious_for_taxon'

    #     for term in ont.keys():
    #         if term.split(':')[0] not in valid_terms:
    #             continue




from setuptools import setup, find_packages
from distutils.util import convert_path

long_description = open('README.md', 'r').read()

# get version:
ns = {}
ver_path = convert_path('ontolopy/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), ns)

setup(name='ontolopy',
      version=ns['__version__'],
      install_requires=[
            'pandas',
            'numpy',
            'validators',
      ],
      description='Ontolopy is a package for working with ontology (.obo) files from Python.',
      long_description_content_type='text/markdown',
      long_description=long_description,
      packages=find_packages(),
      url='https://github.com/NatalieThurlby/ontolopy',
      author='Natalie Thurlby',
      classifiers=[
            "Programming Language :: Python :: 3",
      ],
      python_requires='>=3.6',
      author_email='natalie.thurlby@bristol.ac.uk',
      license='MIT')
